
#include "ArmExtractor.h"

using namespace std;

static unsigned int doLabeling(IplImage* src, unsigned int* labelMap);

static void fillArea( IplImage * img, int x0, int y0, int c ) {
	int w = img->width, h=img->height;

	vector<int> list;
	list.clear( );

	list.push_back( x0 );
	list.push_back( y0 );
	
	int c0 = img->imageData[ (x0)*img->widthStep + y0 ];
	
	img->imageData[ (x0)*img->widthStep + y0 ] = c;
	
	while( !list.empty( ) ) {
		int y = list.back( ); list.pop_back( );
		int x = list.back( ); list.pop_back( );
		
		if ( x > 0 &&  img->imageData[ (x-1)*img->widthStep + y ] == c0  ) {
			img->imageData[ (x-1)*img->widthStep + y ] = c;
			list.push_back( x-1 );
			list.push_back( y );
		}
		if ( y > 0 &&  img->imageData[ (x)*img->widthStep + y-1 ] == c0  ) {
			img->imageData[ (x)*img->widthStep + y-1 ] = c;
			list.push_back( x );
			list.push_back( y-1 );
		}
		if ( x < h-1 &&  img->imageData[ (x+1)*img->widthStep + y ] == c0  ) {
			img->imageData[ (x+1)*img->widthStep + y ] = c;
			list.push_back( x+1 );
			list.push_back( y );
		}
		if ( y < w-1 &&  img->imageData[ (x)*img->widthStep + y+1 ] == c0  ) {
			img->imageData[ (x)*img->widthStep + y+1 ] = c;
			list.push_back( x );
			list.push_back( y+1 );
		}
	}
}

void changeColor( IplImage * img, int sc, int dc ) {
	int w = img->width, h=img->height;
	
	for( int i = 0; i < h; i++ )
		for( int j = 0; j < w; j++ )
			if( img->imageData[ i*img->widthStep+j ] == sc )
				img->imageData[ i*img->widthStep+j ] = dc;

}

static void handFilling( IplImage * img ) {
	int i, j;
	int w = img->width, h=img->height;

	if ( img->imageData[ 0 ] == 0 ) {
		fillArea( img, 0, 0, 127 );
	}
	if ( img->imageData[ w-1 ] == 0 ) {
		fillArea( img, 0, w-1, 127 );
	}
	if ( img->imageData[ (h-1)*img->widthStep ] == 0 ) {
		fillArea( img, h-1, 0, 127 );
	}
	if ( img->imageData[ (h-1)*img->widthStep + w-1 ] == 0 ) {
		fillArea( img, h-1, w-1, 127 );
	}

	changeColor( img, 0, 255 );
	changeColor( img, 127, 0 );
}


void ArmExtractor::extract( IplImage * img, IplImage * hand ) {
	int w = img->width;
	int h = img->height;
	
	unsigned int labelMap[ w*h ];
	doLabeling( img, labelMap );
	
	unsigned int maxLabel = 0;
	for( int i = 0; i < h; i++ )
		for( int j = 0; j < w; j++ )
			if ( labelMap[i*w+j] > maxLabel )
				maxLabel = labelMap[i*w+j];
				
	
	if ( maxLabel < 1 ) {
		for( int i = 0; i < h; i++ )
			for( int j = 0; j < w; j++ ) {
				int offset = i*hand->widthStep + j ;
				
				hand->imageData[ offset + 0 ] = 0;
			
			}
		return;
	}
	
	// Finding the largest region
	int largestSize = 0;
	int largestLabel = 0;
	
	for( int l = 1; l <= maxLabel; l++ ) {
		int size = 0;
		for( int i = 0; i < h; i++ )
			for( int j = 0; j < w; j++ )
				if ( labelMap[i*w+j] == l )
					++size;
					
		if( size > largestSize ) {
			largestSize = size;
			largestLabel = l;
		}
	}
	
	// Extract the hand
	for( int i = 0; i < h; i++ )
		for( int j = 0; j < w; j++ ) {
			int offset = i*hand->widthStep + j ;
			
			if ( labelMap[i*w+j] == largestLabel ) {
				hand->imageData[ offset + 0 ] = 255;
			} else {
				hand->imageData[ offset + 0 ] = 0;
			}
		
		}
	hand->origin = img->origin;
	
	handFilling( hand );
}


void ArmExtractor::extract( IplImage * img, IplImage * hand1, IplImage * hand2 ) {
	int w = img->width;
	int h = img->height;
	
	unsigned int labelMap[ w*h ];
	doLabeling( img, labelMap );
	
	unsigned int maxLabel = 0;
	for( int i = 0; i < h; i++ )
		for( int j = 0; j < w; j++ )
			if ( labelMap[i*w+j] > maxLabel )
				maxLabel = labelMap[i*w+j];
				
	
	if ( maxLabel < 2 ) {
		for( int i = 0; i < h; i++ )
			for( int j = 0; j < w; j++ ) {
				int offset = i*hand1->widthStep + j ;
				
				hand1->imageData[ offset + 0 ] = 0;
				hand2->imageData[ offset + 0 ] = 0;
			
			}
		return;
	}
	
	// Finding the largest region
	int largestSize1 = 0;
	int largestSize2 = 0;
	int largestLabel1 = 0;
	int largestLabel2 = 0;
	
	for( int l = 1; l <= maxLabel; l++ ) {
		int size = 0;
		for( int i = 0; i < h; i++ )
			for( int j = 0; j < w; j++ )
				if ( labelMap[i*w+j] == l )
					++size;
					
		if( size > largestSize1 ) {
			largestSize2 = largestSize1;
			largestSize1 = size;

			largestLabel2 = largestLabel1;
			largestLabel1 = l;
		} else
		if( size > largestSize2 ) {
			largestSize2 = size;
			largestLabel2 = l;
		}
	}
	
	
	// Extract the first hand
	for( int i = 0; i < h; i++ )
		for( int j = 0; j < w; j++ ) {
			int offset = i*hand1->widthStep + j ;
			
			if ( labelMap[i*w+j] == largestLabel1 ) {
				hand1->imageData[ offset + 0 ] = 255;
			} else {
				hand1->imageData[ offset + 0 ] = 0;
			}
		
		}

	// Extract the second hand
	for( int i = 0; i < h; i++ )
		for( int j = 0; j < w; j++ ) {
			int offset = i*hand1->widthStep + j ;
			
			if ( labelMap[i*w+j] == largestLabel2 ) {
				hand2->imageData[ offset + 0 ] = 255;
			} else {
				hand2->imageData[ offset + 0 ] = 0;
			}
		
		}
	hand1->origin = img->origin;
	hand2->origin = img->origin;
	
	float avgX1 = 0;
	float avgX2 = 0;
	for( int i = 0; i < h; i++ )
		for( int j = 0; j < w; j++ ) {
			if ( labelMap[i*w+j] == largestLabel1  )
				avgX1 += j;
			if ( labelMap[i*w+j] == largestLabel2  )
				avgX2 += j;
		}
	avgX1 /= largestSize1;
	avgX2 /= largestSize2;
	
	if ( avgX1 > avgX2 ) {
		IplImage *temp = cvCreateImage( cvSize( img->width, img->height ), IPL_DEPTH_8U, 1 );
		
		cvCopy( hand1, temp );
		cvCopy( hand2, hand1 );
		cvCopy( temp, hand2 );
		
		cvReleaseImage( &temp );
	}
	handFilling( hand1 );
	handFilling( hand2 );
}


static void pointCounter(unsigned int* labelMap, int w, int h, unsigned int labels, unsigned long *count){

    for (int i=0; i<labels; i++){
        count[i]=0;
    }

    for (int i=0; i<h; i++){
        for(int j=0; j<w; j++){
            if (labelMap[i*w+j]) {
                count[labelMap[i*w+j]-1]++;
            }
        }
    }
}


static unsigned int doLabeling(IplImage* src, unsigned int* labelMap){
     int w = src->width;
     int h = src->height;
     unsigned int labels;

     for (int i=0; i<w*h; i++) labelMap[i]=0;

     map<unsigned int, unsigned int> connectedLabels;

     labels=0;

     unsigned int n[4];
     for (int i=0; i<h; i++){
         for (int j=0; j<w; j++){
             if (
		(
		  src->nChannels==1 && (unsigned char)src->imageData[i*src->widthStep+j]
		) ||
		(src->nChannels==3 && 
			(
			  (unsigned char)src->imageData[i*src->widthStep+j*3+0] ||
			  (unsigned char)src->imageData[i*src->widthStep+j*3+1] ||
			  (unsigned char)src->imageData[i*src->widthStep+j*3+2]
			)
		)
	     ){
                 n[0]=n[1]=n[2]=n[3]=0;
                 if (j>0) n[0]= labelMap[i*w+j-1];
                 if (i>0 && j>0) n[1]= labelMap[i*w+j-w-1];
                 if (i>0) n[2]= labelMap[i*w+j-w];
                 if (i>0 && j<w-1) n[3]= labelMap[i*w+j-w+1];

                 int cnt=0;
                 for (int k=0; k<4; k++) if (n[k]) cnt++;


                 if (!cnt){
                    labels++;
                    labelMap[i*w+j] = labels;
                 }
                 else if (cnt==1){
                    labelMap[i*w+j] = n[0] ? n[0] : n[1] ? n[1] : n[2] ? n[2] : n[3];
                 }
                 else if (cnt>1) {

                      if (n[0] && n[3]) {
                         unsigned int min, max;

                         if (n[0]<n[3]) { min = n[0]; max=n[3]; }
                         else           { min = n[3]; max=n[0]; }


                         map<unsigned int,unsigned int>::iterator it;
                         bool enabled=true;
                         if((it=connectedLabels.find(max)) != connectedLabels.end()){
                           if(it->second < min) {
                             enabled=false;
                             min=it->second;
                           }
                         }

                         labelMap[i*w+j] = min;
                         if (min!=max && enabled) {
                            connectedLabels.insert(pair<unsigned int, unsigned int>(max, min));
                         }

                      }
                      else if (n[1] && n[3]) {
                         unsigned int min, max;
                         if (n[1]<n[3]) { min = n[1]; max=n[3]; }
                         else           { min = n[3]; max=n[1]; }

                         labelMap[i*w+j] = min;

                         map<unsigned int,unsigned int>::iterator it;
                         bool enabled=true;
                         if((it=connectedLabels.find(max)) != connectedLabels.end()){
                           if(it->second > min) enabled=false;
                         }

                         if (min!=max && enabled) connectedLabels.insert(pair<unsigned int, unsigned int>(max, min));
                      }
                      else {
                         unsigned int min=n[0];
                         for (int k=0; k<4; k++) min = ((min>=n[k] && n[k]) || !min) ? n[k] : min;
                         labelMap[i*w+j] = min;

                      }

                 }

             }
             else {
                  labelMap[i*w+j] = 0;
             }
         }


     }

     map<unsigned int,unsigned int>::reverse_iterator rit;
     map<unsigned int,unsigned int>::iterator it;
     unsigned int col;
     for (rit=connectedLabels.rbegin(); rit != connectedLabels.rend(); rit++){
         col=rit->second;
         while((it=connectedLabels.find(col))!=connectedLabels.end()){
            col=it->second;
         }

         rit->second = col;

     }

     for(int i=0; i<h; i++){
        for (int j=0; j<w; j++){
            if ((it = connectedLabels.find(labelMap[i*w+j])) != connectedLabels.end() ){
               labelMap[i*w+j]= it->second;
            }
        }
     }



     unsigned long count[labels];
     pointCounter(labelMap, w, h, labels, count);
     unsigned int actLabel=1;
     unsigned int lastLabel=labels;
     while(actLabel<lastLabel){

            while(!count[lastLabel-1] && lastLabel) lastLabel--;

            if (!count[actLabel-1] && (actLabel<lastLabel)){
                 for(int i=0; i<h; i++){
                    for (int j=0; j<w; j++){
                           if (labelMap[i*w+j]==lastLabel) labelMap[i*w+j]=actLabel;
                    }
                 }

                 count[actLabel-1]=count[lastLabel-1];
                 count[lastLabel-1]=0;

                 lastLabel--;
            }

        actLabel++;
     }
     while(!count[lastLabel-1] && lastLabel) lastLabel--;

     return lastLabel;
}
 
