#include "HandRecognition.h"

#define HU_MIN 1.0E-27

void HuMoments(IplImage *input, CvMoments *moments, CvHuMoments *hu_moments) {

  cvMoments(input,moments,0);
  cvGetHuMoments(moments,hu_moments);

  if (hu_moments->hu1<HU_MIN) hu_moments->hu1=HU_MIN;
  if (hu_moments->hu2<HU_MIN) hu_moments->hu2=HU_MIN;
  if (hu_moments->hu3<HU_MIN) hu_moments->hu3=HU_MIN;
  if (hu_moments->hu4<HU_MIN) hu_moments->hu4=HU_MIN;
  if (hu_moments->hu5<HU_MIN) hu_moments->hu5=HU_MIN;
  if (hu_moments->hu6<HU_MIN) hu_moments->hu6=HU_MIN;
  if (hu_moments->hu7<HU_MIN) hu_moments->hu7=HU_MIN;
}

static const char *hand_types[]={"    ?    ","   Rock   "," Scissors ","   Paper  ","Paper2"};

const char *handType(RSPtype what) { int tmp=(int)what; return (tmp>=0&&tmp<5)?hand_types[tmp]:NULL; }

HandRecognition::HandRecognition(int _hist_len) {
  int i;

  hist_len=_hist_len; if (hist_len>HR_MAX_HIST) hist_len=HR_MAX_HIST;

  hist_pos=0; for (i=0;i<hist_len;i++) hist[i]=rspUnknown;
}

HandRecognition::~HandRecognition(void) { classifier.clear(); }

bool HandRecognition::train(const char *fn) {
  bool result=FALSE;
  int i, j, n, cl;
  double hu;
  FILE *f;

  f=fopen(fn,"r");

  if (f!=NULL) {

    fscanf(f,"%d",&n);

    CvMat *samples=cvCreateMat(n,3,CV_32F); //7
    CvMat *responses=cvCreateMat(n,1,CV_32F);

    for (i=0;i<n;i++) {
      fscanf(f,"%d:",&cl);
      CV_MAT_ELEM(*responses,float,i,0)=cl;

      for (j=0;j<7;j++) {
        fscanf(f,"%lf",&hu);
        if (j>0&&j<4) CV_MAT_ELEM(*samples,float,i,j-1)=hu; //
      }
    }

    fclose(f);

    result=classifier.train(samples,responses);

    cvReleaseMat(&samples);
    cvReleaseMat(&responses);
  }

  return result;
}

RSPtype HandRecognition::classify(IplImage *input) {
  float cl;

  CvMoments moments;
  CvHuMoments hu_moments;

  HuMoments(input,&moments,&hu_moments);

  CvMat *sample=NULL;

  sample=cvCreateMat(1,3,CV_32FC1); //7
  //CV_MAT_ELEM(*sample,float,0,0)=log(hu_moments.hu1);
  CV_MAT_ELEM(*sample,float,0,1-1)=log(hu_moments.hu2);
  CV_MAT_ELEM(*sample,float,0,2-1)=log(hu_moments.hu3);
  CV_MAT_ELEM(*sample,float,0,3-1)=log(hu_moments.hu4);
  //CV_MAT_ELEM(*sample,float,0,4)=log(hu_moments.hu5);
  //CV_MAT_ELEM(*sample,float,0,5)=log(hu_moments.hu6);
  //CV_MAT_ELEM(*sample,float,0,6)=log(hu_moments.hu7);

  cl=classifier.predict(sample);

  cvReleaseMat(&sample);

  return (RSPtype)((int)cl);
}

RSPtype HandRecognition::history(RSPtype choose, float *prob) {
  int max, clbin[rspEnd], i, j;
  RSPtype maxcl;

  hist[hist_pos++]=choose; if (hist_pos==hist_len) hist_pos=0;

  for (j=(int)rspUnknown;j<(int)rspEnd;j++) clbin[j]=0;

  for (i=0;i<hist_len;i++) clbin[hist[i]]++;

  max=-1; maxcl=rspUnknown;
  
  for (j=(int)rspUnknown;j<(int)rspEnd;j++) {
    if (max<clbin[j]) {
      max=clbin[j]; maxcl=(RSPtype)j;
    } else if (max==clbin[j]) maxcl=rspUnknown;
  }

  if (prob) *prob=max/(float)hist_len;

  return maxcl;
}

int computeHuMoments(char *srcDir, char *outPath) {
  WIN32_FIND_DATA FileData;
  char tmp[MAX_PATH];
  HANDLE hSearch;
  FILE *f;

  f=fopen(outPath,"w");

  strcpy(tmp,srcDir); strcat(tmp,"*.png");
  hSearch=FindFirstFile(tmp,&FileData);

  if ((f!=NULL)&&(hSearch!=INVALID_HANDLE_VALUE)) do {

    strcpy(tmp,srcDir);
    strcat(tmp,FileData.cFileName);

    IplImage *input=cvLoadImage(tmp,CV_LOAD_IMAGE_GRAYSCALE);

    int i, j, l, k, k1, k2, size=min(input->width,input->height), half=size/2;
    int dx1, dy1, dx2, dy2, cnt1, cnt2;
    char *ptr, *ptrl;

    {
      CvMoments moments;
      CvHuMoments hu_moments;

      HuMoments(input,&moments,&hu_moments);

      fprintf(f,"%s: % 6.3f % 6.3f % 6.3f % 6.3f % 6.3f % 6.3f % 6.3f\n",FileData.cFileName,log(hu_moments.hu1),log(hu_moments.hu2),log(hu_moments.hu3),log(hu_moments.hu4),log(hu_moments.hu5),log(hu_moments.hu6),log(hu_moments.hu7));
    }

    if (input) cvReleaseImage(&input);

    if ((cvWaitKey()&255)==27) break;

  } while (FindNextFile(hSearch,&FileData));

  fclose(f);

  return 0;
}

int classifyDir(char *srcDir, char *outPath, char *trainset) {
  WIN32_FIND_DATA FileData;
  char tmp[MAX_PATH];
  HANDLE hSearch;
  FILE *f;

  f=fopen(outPath,"w");

  strcpy(tmp,srcDir); strcat(tmp,"*.png");
  hSearch=FindFirstFile(tmp,&FileData);

  HandRecognition hr;

  hr.train(trainset);

  if ((f!=NULL)&&(hSearch!=INVALID_HANDLE_VALUE)) do {

    strcpy(tmp,srcDir);
    strcat(tmp,FileData.cFileName);

    IplImage *input=cvLoadImage(tmp,CV_LOAD_IMAGE_GRAYSCALE);

//    handFilling(input); cutHand(input);
    fprintf(f,"%s: %12s\n",FileData.cFileName,hand_types[hr.classify(input)]);

    if (input) cvReleaseImage(&input);

    if ((cvWaitKey()&255)==27) break;

  } while (FindNextFile(hSearch,&FileData));

  fclose(f);

  return 0;
}
