
#include "recognizer.h"

#define MAX_VALUE 1000000

//--------------------------------------------------------


void Recognizer::contour(IplImage* src, IplImage* dest){
    for(int i=1; i<dest->height-1; i++){
        for(int j=1; j<dest->width-1; j++){
            bool contour = false;
            if( (unsigned int)src->imageData[dest->widthStep*i + j] > 0) {
                for(int y=-1; y<2; y++){
                    for(int x=-1; x<2; x++){
                        int offset = dest->widthStep*(i+y) + j + x;
                        if((unsigned int)src->imageData[offset] == 0){
                            contour = true;
                            break;
                        }

                    }
                    if (contour) break;
                }
            }

            if (contour){
                dest->imageData[dest->widthStep*i + j] = (unsigned char)255;
            }
            else {
                dest->imageData[dest->widthStep*i + j] = (unsigned char)0;
//DEBUG
//                dest->imageData[dest->widthStep*i + j] = (unsigned char)src->imageData[dest->widthStep*i + j];
            }
        }
    }
}

void drawDistanceMap(unsigned int* dm, IplImage* dest){

    for(int i=0; i<dest->height; i++){
        for(int j=0; j<dest->width; j++){
            if (dm[i*dest->width+j]<MAX_VALUE)
                dest->imageData[i*dest->widthStep+j]=dm[i*dest->width+j];
            else
                dest->imageData[i*dest->widthStep+j]=0;

        }
    }
}

void Recognizer::getDistanceMap(IplImage* src, unsigned int* dm){
    for(int i=0; i<src->height*src->width; i++) dm[i]=MAX_VALUE;

    for(int i=0; i<src->height; i++){
        for(int j=0; j<src->width; j++){
            if(!(unsigned int)src->imageData[i*src->widthStep + j]) dm[i*src->width+j]=0;

            unsigned int min = dm[i*src->width+j];
            if(i-1>=0){
                if (dm[(i-1)*src->width+j] < min) min = dm[(i-1)*src->width+j];
                if(j-1>=0){
                    if (dm[(i-1)*src->width+j-1] < min) min = dm[(i-1)*src->width+j-1];
                }
                if(j+1<src->width){
                    if (dm[(i-1)*src->width+j+1] < min) min = dm[(i-1)*src->width+j+1];
                }
            }
            if(i+1<src->height){
    //            if (dm[(i+1)*src->width+j] < min) min = dm[(i+1)*src->width+j];
                if(j-1>=0){
                    if (dm[(i+1)*src->width+j-1] < min) min = dm[(i+1)*src->width+j-1];
                }
   /*             if(j+1<src->width){
                    if (dm[(i+1)*src->width+j+1] < min) min = dm[(i+1)*src->width+j+1];
                }
   */
            }

            if(j-1>=0){
                if (dm[i*src->width+j-1] < min) min = dm[i*src->width+j-1];
            }
            if(j+1<src->width){
                if (dm[i*src->width+j+1] < min) min = dm[i*src->width+j+1];
            }

            if (min< dm[i*src->width+j]) {
                dm[i*src->width+j] = min + 1;
            }
        }
    }

    for(int i=src->height-1; i>=0; i--){
        for(int j=src->width-1; j>=0; j--){
            if(!(unsigned int)src->imageData[i*src->widthStep + j]) dm[i*src->width+j]=0;

            unsigned int min = dm[i*src->width+j];
            if(i-1>=0){
      /*          if (dm[(i-1)*src->width+j] < min) min = dm[(i-1)*src->width+j];
                if(j-1>=0){
                    if (dm[(i-1)*src->width+j-1] < min) min = dm[(i-1)*src->width+j-1];
                }
       */
                if(j+1<src->width){
                    if (dm[(i-1)*src->width+j+1] < min) min = dm[(i-1)*src->width+j+1];
                }
            }

            if(i+1<src->height){
                if (dm[(i+1)*src->width+j] < min) min = dm[(i+1)*src->width+j];
                if(j-1>=0){
                    if (dm[(i+1)*src->width+j-1] < min) min = dm[(i+1)*src->width+j-1];
                }
                if(j+1<src->width){
                    if (dm[(i+1)*src->width+j+1] < min) min = dm[(i+1)*src->width+j+1];
                }
            }

            if(j-1>=0){
                if (dm[i*src->width+j-1] < min) min = dm[i*src->width+j-1];
            }
            if(j+1<src->width){
                if (dm[i*src->width+j+1] < min) min = dm[i*src->width+j+1];
            }

            if (min< dm[i*src->width+j]) {
                dm[i*src->width+j] = min + 1;
            }
        }
    }
}





void Recognizer::getBoundingBox(IplImage* src, int& x1, int& y1, int& x2, int& y2){
    x2=y2=0;
    x1=src->width-1;
    y1=src->height-1;
    for(int i=1; i<src->height; i++){
        for(int j=0; j<src->width; j++){
            if((unsigned char)src->imageData[i*src->widthStep+j]!=0){
                if(i<y1) y1=i;
                if(i>y2) y2=i;
                if(j<x1) x1=j;
                if(j>x2) x2=j;
            }
        }        
    }
}


unsigned long Recognizer::getAera(IplImage* src){
    unsigned long aera=0;
    for (int i=0; i<src->height; i++){
        for(int j=0; j<src->width; j++){
            if((unsigned char)src->imageData[i*src->widthStep+j]!=0){
                aera++;
            }
        }
    }
    
    return aera;
}

unsigned long Recognizer::getBoundary(IplImage* src){
    unsigned long boundary=0;

    for (int i=0; i<src->height; i++){
        for(int j=0; j<src->width; j++){
            unsigned char p = src->imageData[i*src->widthStep+j];
            if (p) {
                if (j>0){
                    if(!(unsigned char)src->imageData[i*src->widthStep+j-1]) boundary++;
                }
                if (i>0){
                    if(!(unsigned char)src->imageData[(i-1)*src->widthStep+j]) boundary++;
                }
                if (j<src->width-1){
                    if(!(unsigned char)src->imageData[i*src->widthStep+j+1]) boundary++;
                }
                if (i<src->height-1){
                    if(!(unsigned char)src->imageData[(i+1)*src->widthStep+j]) boundary++;
                }

            }
        }
    }
    
    return boundary;
}



//------------------------------------------------------------------------------------

Recognizer::Recognizer(){
}

void Recognizer::getShapeStat(IplImage* src, unsigned long* stat){
    
}


void Recognizer::cutHand(IplImage *src){

    unsigned long act=0;
    unsigned long prev=0;
    int maxpos=0;
    int cutpos=src->width;
    unsigned int max=0;

    for(int j=0; j<src->width; j++){
        act=0;
        for(int i=0; i<src->height; i++){
            if((unsigned char)src->imageData[i*src->widthStep+j]!=0) act++;
        }

        if(max<act) {
            max=act;
            maxpos=j;
        }

        if(max>10 && (act<(double)max/(double)1.4)) {
            cutpos=j;
            break;
        }

    }

  //  for(int i=0;i<dest->height; i++) dest->imageData[i*dest->widthStep + maxpos] = (unsigned char)128;
  //  for(int i=0;i<dest->height; i++) dest->imageData[i*dest->widthStep + cutpos] = (unsigned char)128;


    for(int i=0; i<src->height; i++){
        for(int j=cutpos; j<src->width; j++){
            src->imageData[i*src->widthStep + j] = (unsigned char)0;
        }
    }


}

#include <cstdlib>
#include <iostream>
int Recognizer::recognizeHandSignal(IplImage* src){

    
    int y1, x1, x2, y2;
    getBoundingBox(src,x1,y1,x2,y2);
    
    for(int i=x1;i<=x2; i++) src->imageData[y1*src->widthStep + i] = (unsigned char)128;
    for(int i=x1;i<=x2; i++) src->imageData[y2*src->widthStep + i] = (unsigned char)128;
    for(int i=y1;i<=y2; i++) src->imageData[i*src->widthStep + x1] = (unsigned char)128;
    for(int i=y1;i<=y2; i++) src->imageData[i*src->widthStep + x2] = (unsigned char)128;
    
    unsigned long aera     = getAera(src);
    unsigned long boundary = getBoundary(src);
    
//    double density = (double)aera/(double)((x2-x1)*(y2-y1));
    double density = (double)aera/(double)boundary;

//    printf("%f\n", density);
    if(fabs((double)(x2-x1)/(double)(y2-y1) - 1.0)<0.25) {
        return ROCK;
    }
    else if(density>=0.7){
        return SCISSORS;
    }
    else {
        return PAPER;
    }
    
    
    return UNKNOWN;
}
